
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.7.0rc1'
version = '0.7.0rc1'
git_revision = '7dcc9b59cc6b96ad20114bbe7566697e81dd54ac'
full_version = '0.7.0rc1.dev-%(git_revision)s' % {
        'git_revision': git_revision}
release = True

if not release:
    version = full_version
